"""
Copyright 2021 by Sergei Belousov
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import typing
import torch
import torch.nn as nn
import torch.nn.functional as F


class Normalize(nn.Module):
    """ Implementation of differentiable normalization of the input tensor
    Arguments:
        mean (list<float>): per channel mean.
        std (list<float>): per channel mean.
        input_range (tuple<float, float>): range of the input values.

    """
    def __init__(
            self,
            mean: typing.List[float],
            std: typing.List[float],
            input_range: typing.Tuple[float, float] = (-1.0, 1.0)
    ):
        super().__init__()
        # input range
        self.input_range = input_range
        # project input_range -> [0, 1]
        range_shift = input_range[0]
        range_scale = input_range[1] - input_range[0] + 1e-5
        # prepare mean
        mean = torch.Tensor(mean).view(1, -1, 1, 1)
        mean = range_shift + range_scale * mean
        self.register_buffer("mean", mean)
        # prepare std
        std = torch.Tensor(std).view(1, -1, 1, 1)
        std = range_scale * std
        self.register_buffer("std", std)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Normalize input tensor.
        Arguments:
            x (torch.Tensor): input tensor.
        Returns:
            y (torch.Tensor): normalized tensor.
        """
        # clamp range
        x = torch.clamp(x, self.input_range[0], self.input_range[1])
        # normalize
        x = (x - self.mean) / (self.std + 1e-5)
        return x


class Resize(nn.Module):
    """ Implementation of differentiable resize of the input tensor
    Arguments:
        size (int or tuple<int, int>): output size.
        mode (str): interpolation mode (default = "bilinear").

    """
    def __init__(self, size: int or typing.Tuple[int, int], mode: str = "bilinear"):
        super().__init__()
        self.size = size if isinstance(size, tuple) else (size, size)
        self.mode = mode

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Resize input tensor.
        Arguments:
            x (torch.Tensor): input tensor.
        Returns:
            y (torch.Tensor): resized tensor.
        """
        if x.size(2) == self.size[0] and x.size(3) == self.size[1]:
            return x
        return F.interpolate(x, size=self.size, mode=self.mode)



"""
Copyright 2021 by Sergei Belousov
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os


def get_module_path():
    """ Get module path
    Returns:
        path (str): path to current module.
    """
    file_path = os.path.abspath(__file__)
    module_path = os.path.dirname(file_path)
    return module_path


def res_path(path):
    """ Resource path
    Arguments:
        path (str): related path from module dir to some resources.
    Returns:
        path (str): absolute path to module dir.
    """
    return os.path.join(get_module_path(), path)



"""
Copyright 2021 by Sergei Belousov
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import typing
import torch
import youtokentome as yttm
from transformers import CLIPTokenizer
from googletrans import Translator
from .utils import tokens_to_tensor


class GoogleTranslate:
    """ Implementation of the API to translate.google.com
    Arguments:
        target_lang (str): target language.
    """
    def __init__(self, target_lang: str = 'ru'):
        self.target_lang = target_lang
        self.translator = Translator()

    def __call__(self, text: str):
        """ Translate input text to target language if needs.
        Arguments:
            text (str): input text.
        Returns:
            translation (str): output text in target language.
        """
        try:
            source_lang = self.translator.detect(text).lang
            if source_lang != self.target_lang:
                print(f"translate {source_lang}->{self.target_lang}")
                translation = self.translator.translate(text, dest=self.target_lang).text
            else:
                translation = text
        except:
            translation = text
        return translation


class TextProcessor:
    """ Base class for text processor.
    Arguments:
        target_lang (str): target language.
    """
    def __init__(self, target_lang: typing.Optional[str] = None):
        self.translator = GoogleTranslate(target_lang) if target_lang is not None else None

    def __len__(self) -> int:
        """ Return vocabulary size.
        Returns:
            length (int): vocabulary size.
        """
        return self.vocab_size()

    def vocab_size(self) -> int:
        """ Return vocabulary size.
        Returns:
            length (int): vocabulary size.
        """
        raise NotImplemented

    def encode(
            self,
            text: str,
            *args, **kwargs
    ) -> typing.Dict[str, torch.Tensor]:
        """ Encode input text.
        Arguments:
            text (str): input text.
            *args, **kwargs: auxiliary arguments.
        Returns:
            encoded_text (dict<str:torch.Tensor>): encoded text.
        """
        raise NotImplemented

    def decode(self, tokens: torch.Tensor) -> str:
        """ Decode input tokens.
        Arguments:
            tokens (torch.Tensor): input tokens.
        Returns:
            text (str): decoded text.
        """
        raise NotImplemented

    def translate(self, text: str) -> str:
        """ Translate text if needed.
        Arguments:
            text (str): input text.
        Returns:
            text (str): output text.
        """
        if self.translator is not None:
            text = self.translator(text)
        return text


class YTTMTokenizerTextProcessor(TextProcessor):
    """ Implementation of the YTTM tokenizer.
    Arguments:
        model_path (str): path to pretrained model.
        target_length (int): target length of the encoded tokens.
        target_lang (str): target language.
    """
    eos_id = 3
    bos_id = 2
    unk_id = 1
    pad_id = 0

    def __init__(self, model_path: str, target_length=None, target_lang=None):
        super().__init__(target_lang)
        self.target_length = target_length
        self.tokenizer = yttm.BPE(model=model_path)

    def vocab_size(self) -> int:
        """ Return vocabulary size.
        Returns:
            length (int): vocabulary size.
        """
        return self.tokenizer.vocab_size()

    def encode(
            self,
            text: str,
            to_lowercase: bool = True,
            dropout_prob: float = 0.0,
            return_mask: bool = False
    ) -> typing.Dict[str, torch.Tensor]:
        """ Encode input text.
        Arguments:
            text (str): input text.
            to_lowercase (bool): edit text to lowercase.
            dropout_prob (float): dropout prob.
            return_mask (bool): return attention mask
        Returns:
            encoded_text (dict<str:torch.Tensor>): encoded text.
        """
        text = self.translate(text)
        if to_lowercase:
            text = text.lower()
        tokens = self.tokenizer.encode(
            [text.strip()],
            output_type=yttm.OutputType.ID,
            dropout_prob=dropout_prob
        )[0]
        tokens = [self.bos_id] + tokens + [self.eos_id]
        if self.target_length is None:
            target_length = len(tokens)
        else:
            target_length = self.target_length
        output = {}
        output["input_ids"] = tokens_to_tensor(tokens, target_length, self.pad_id)
        if return_mask:
            output["attention_mask"] = (output["input_ids"] != self.pad_id)
        return output

    def decode(self, tokens: torch.Tensor) -> str:
        """ Decode input tokens.
        Arguments:
            tokens (torch.Tensor): input tokens.
        Returns:
            text (str): decoded text.
        """
        return self.tokenizer.decode(
            tokens.cpu().numpy().tolist(),
            ignore_ids=[self.eos_id, self.bos_id, self.unk_id, self.pad_id]
        )[0]


class HFCLIPTextProcessor(TextProcessor):
    """ Implementation of the Huggingface CLIP tokenizer.
    Arguments:
        tokenizer (CLIPTokenizer): pre-trained CLIP tokenizer.
        target_lang (str): target language.
    """
    def __init__(self, tokenizer: CLIPTokenizer, target_lang: typing.Optional[str] = None):
        super().__init__(target_lang)
        self.tokenizer = tokenizer

    def vocab_size(self) -> int:
        """ Return vocabulary size.
        Returns:
            length (int): vocabulary size.
        """
        return self.tokenizer.vocab_size()

    def encode(
            self,
            text: str,
            return_tensors: str = "pt",
            return_mask: bool = True,
            **kwargs
    ) -> typing.Dict[str, torch.Tensor]:
        """ Encode input text.
        Arguments:
            text (str): input text.
            return_tensor (str): return tokens as torch.Tensor.
            return_mask (bool): return attention mask
            **kwargs: auxiliary arguments.
        Returns:
            encoded_text (dict<str:torch.Tensor>): encoded text.
        """
        text = self.translate(text)
        output = self.tokenizer(text, return_tensors=return_tensors, **kwargs)
        if not return_mask:
             del output["attention_mask"]
        return output

    def decode(
            self,
            tokens: torch.Tensor,
            skip_special_tokens: bool = True,
            clean_up_tokenization_spaces: bool = True
    ) -> str:
        """ Decode input tokens.
        Arguments:
            tokens (torch.Tensor): input tokens.
            skip_special_tokens (bool): clean special tokens.
            clean_up_tokenization_spaces (bool): clean special tokens.
        Returns:
            text (str): decoded text.
        """
        if tokens.ndim == 1:
            return self.tokenizer.decode(
                tokens,
                skip_special_tokens=skip_special_tokens,
                clean_up_tokenization_spaces=clean_up_tokenization_spaces
            )
        else:
            return self.tokenizer.decode_batch(
                tokens,
                skip_special_tokens=skip_special_tokens,
                clean_up_tokenization_spaces=clean_up_tokenization_spaces
            )



"""
Copyright 2021 by Sergei Belousov
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import typing
import torch.nn as nn
from omegaconf import OmegaConf
from .utils import res_path
from .models import CLIP, ruCLIP
from .processor import TextProcessor


def get_models_list() -> typing.List[str]:
    """ Get list of available CLIP models.
    Returns:
        models_list (list[str]): list of available models.
    """
    return [name.split(".")[0] for name in os.listdir(res_path(f"configs"))]


def get_clip_model(clip_type: str = "clip_vit_b32", input_range: typing.Tuple[float, float] = (-1.0, 1.0), cache_dir: str = "/tmp/") -> typing.Tuple[nn.Module, TextProcessor, nn.Module]:
    """Build CLIP model from config file.
    Arguments:
        cfg (OmegaConf): configuration of the model.
        input_range (tuple[float, float]): input range.
        cache_dir (str): path to cache dir.
    Returns:
        model (nn.Module): CLIP model.
        text_processor (TextProcessor): text processor.
        image_processor (nn.Module): image processor.
    """
    assert clip_type in get_models_list(), f"Unknown clip_type: {clip_type}"
    cfg = OmegaConf.load(res_path(f"configs/{clip_type}.yml"))
    if cfg.type == "clip":
        return CLIP.from_pretrained(cfg, input_range, cache_dir=cache_dir)
    elif cfg.type  == "ruclip":
        return ruCLIP.from_pretrained(cfg, input_range, cache_dir=cache_dir)
    else:
        raise ValueError(f"Unknown model type: {cfg.type}. Available model types: [clip, ruclip]")



"""
Copyright 2021 by Sergei Belousov
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import typing
import torch
import torch.nn as nn
from transformers import CLIPModel, CLIPProcessor
from pytorch_clip_guided_loss.processor import HFCLIPTextProcessor, TextProcessor, Resize, Normalize
# utils
from omegaconf import OmegaConf


class CLIP(nn.Module):
    """ Implementation of CLIP model by OpenAI."""
    @staticmethod
    def from_pretrained(
            cfg: OmegaConf,
            input_range: typing.Tuple[float, float] = (-1.0, 1.0),
            cache_dir: str = "/tmp/"
    ) -> typing.Tuple[nn.Module, TextProcessor, nn.Module]:
        """Build model from pre-trained checkpoint.
        Arguments:
            cfg (OmegaConf): configuration of the model.
            input_range (tuple[float, float]): input range.
            cache_dir (str): path to cache dir.
        Returns:
            model (nn.Module): CLIP model.
            text_processor (TextProcessor): text processor.
            image_processor (nn.Module): image processor.
        """
        model = CLIPModel.from_pretrained(cfg.model.config)
        processor = CLIPProcessor.from_pretrained(cfg.processor.config)
        text_processor = HFCLIPTextProcessor(processor.tokenizer, target_lang=cfg.tokenizer.target_lang)
        # load image transforms
        image_processor = nn.Sequential(
            Resize(size=processor.feature_extractor.size),
            Normalize(
                mean=processor.feature_extractor.image_mean,
                std=processor.feature_extractor.image_std,
                input_range=input_range
            )
        )
        return model, text_processor, image_processor



"""
Copyright 2021 by Sergei Belousov
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import typing
import torch.nn as nn
from pytorch_clip_guided_loss.processor import TextProcessor
from omegaconf import OmegaConf
from .clip import CLIP
from .ruclip import ruCLIP


def get_clip_model(cfg: OmegaConf, input_range: typing.Tuple[float, float] = (-1.0, 1.0), cache_dir: str = "/tmp/") -> typing.Tuple[nn.Module, TextProcessor, nn.Module]:
    """Build CLIP model from config file.
    Arguments:
        cfg (OmegaConf): configuration of the model.
        input_range (tuple[float, float]): input range.
        cache_dir (str): path to cache dir.
    Returns:
        model (nn.Module): CLIP model.
        text_processor (TextProcessor): text processor.
        image_processor (nn.Module): image processor.
    """
    if cfg.type == "ruclip":
        return ruCLIP.from_pretrained(cfg, input_range, cache_dir=cache_dir)
    elif cfg.type == "clip":
        return CLIP.from_pretrained(cfg, input_range, cache_dir=cache_dir)
    else:
        raise ValueError(f"Unknown model type: {cfg.type}. Available model types: [clip, ruclip]")



"""
Copyright 2021 by Sergei Belousov

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from setuptools import setup, find_packages

readme = open('README.md').read()

VERSION = '2021.12.7.2'

requirements = [
    'wheel',
    'omegaconf',
    'torch',
    'transformers',
    'huggingface_hub',
    'youtokentome',
    'googletrans',
    'numpy'
]

setup(
    # Metadata
    name='pytorch_clip',
    version=VERSION,
    author='Sergei Belousov aka BeS',
    author_email='sergei.o.belousov@gmail.com',
    description='Pytorch API to work with CLIP models.',
    long_description=readme,
    long_description_content_type='text/markdown',

    # Package info
    packages=find_packages(exclude=('*test*',)),

    #
    zip_safe=True,
    install_requires=requirements,

    # Classifiers
    classifiers=[
        'Programming Language :: Python :: 3',
    ],


    # install .json configs
    package_data={
        "pytorch_clip": ["configs/*.yml"]
    },
    include_package_data=True
)
